// - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
// ZOMBIE TORRENT 2 - GAMEMODE
// Contains game-related things
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

// Basic state machine for the game

enum GameStateType
{
	STATE_LOBBY,				// Game is in the lobby, waiting for players to be ready
	STATE_COUNTDOWN,			// Next wave is starting, we're doing the countdown
	STATE_WAVE,					// The wave has started! Zombies, gameplay, etc.
	STATE_POSTWAVE,				// All zombies are killed, do a short delay to tradertime
	STATE_TRADER,				// We're in tradertime
	STATE_ENDGAME,				// Boss has killed us, or we won!
	STATE_SANDBOX,				// Don't do anything
};

// Holder for gamestate variables
// This keeps it compressed, rather than having a ton of them
//
// (The client has no information about MOST of these settings)

struct GameStateStruct
{
	bool Client;						// Are we on the client side?
	bool Singleplayer;					// Are we playing in singleplayer?
	bool Online;						// Is this a network game?
	bool LobbyOpen;						// SHOULD the lobby menu be open?
	int Spectating;						// Our spectator state

	int LastResX;						// The last known resolution value
	int LastResY;						// The last known resolution value
	
	int State;							// Current state for the game
	int Wave;							// Which wave are we on? (0-based)
	int WaveTimer;						// Wave countdown / trader time
	
	int ZombiePool;						// We can only spawn zombies if this is > 0
	int ZombiesKilled;					// How many zombies have we KILLED so far?
	int ZombieGoal;						// When we kill this many zombies, the round ends
	
	int CurrentTrader;					// Which trader in the array are we using?
	
	struct ZT2HUDInfo HUD;				// HUD information (ZT2HUD.h)
	
	// Info below is set via ZT2UI.h
	str MapTitle;						// People can add colors, etc. to this
	str MapAuthor;						// Who created the map?
	str MapPicture;						// Thumbnail image
	str MapDescription;					// Map description, auto-wrapped
};

struct GameStateStruct GameState;

// Holder for player variables

struct PlayerStateStruct
{
	bool Ready;							// Is this player ready?
	bool Alive;							// This player is currently "alive"
	bool InTrader;						// This player is in the trader
	int TraderIndex;					// WHICH trader are we inside of?
	int PerkIndex;						// THIS MAY BE DIFFERENT ON SERVER AND CLIENT
	int PerkLevel;						// This is for VISUAL INFORMATION ONLY
	
	int ResX;							// Horizontal screen res
	int ResY;							// Vertical screen res
};

struct PlayerStateStruct PlayerState[MAXPLAYERS];

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

script "GameInitServer" OPEN { ZT2Initialize(true); }
script "GameInitClient" OPEN CLIENTSIDE { ZT2Initialize(false); }

// Initialize core game elements!
void ZT2Initialize(int isServer)
{
	// Set clientside variable
	// Checking a bool is better than constantly calculating
	GameState.Client = !isServer;
	
	// Network game?
	GameState.Online = IsNetworkGame();
	
	// Singleplayer?
	GameState.Singleplayer = IsOffline();
	
	// Start syncing resolution-based things
	StartResolutionHook();
	
	RegisterZT2Weapons(isServer);
	RegisterZT2Zombies(isServer);
	RegisterZT2Perks(isServer);
	RegisterZT2Playlist(isServer);

	// Create our persistent HUD
	// This contains things that spectators, etc. should see
	if (!isServer)
	{
		ACS_NamedExecute("CreateZT2HUD", 0, true);
		
		// Request initial gamestate
		ACS_NamedExecute("StateRequest_Wait", 0);
	}
	
	// First joining player should start some cool lobby music
	else if (!Jukebox.HasPlayedInitial)
	{
		int ind = DecideCalmSong();
		SetMusic(Jukebox.custCalm ? Jukebox.MapCalmSongs[ind].Track : Jukebox.CalmSongs[ind].Track);
		Jukebox.HasPlayedInitial = true;
	}
}

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

// Boss wave
bool IsBossWave()
{
	return (GameState.Wave+1 == GetFinalWave());
}

// Final wave, just in case we change this later or during the game
int GetFinalWave()
{
	return MAX_WAVES;
}

// Can the game be IMMEDIATELY joined?
bool GameIsJoinable()
{
	return (GameState.State == STATE_TRADER || GameState.State == STATE_COUNTDOWN);
}

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

// ------------------------/
// Set the time variable for the gamestate
// ------------------------/
void SetGameTimer(int newTime)
{
	GameState.WaveTimer = newTime;
	Rep_UpdateCount();
}

// ------------------------/
// Set the game state
// ------------------------/
void SetGameState(int newState)
{
	GameState.State = newState;
	Rep_UpdateState();
}

// ------------------------/
// From the lobby, start the actual game!
// ------------------------/

void StartInvasionGame()
{
	SetActivator(-1);
	
	// First wave
	GameState.Wave = 0;
	RevivePlayers();
	
	if (!IsSandbox())
		BeginWave();
	else
		SetGameState(STATE_SANDBOX);
}

// ------------------------/
// Actually start a wave!
// ------------------------/

void BeginWave()
{	
	SetMusic("");
	SetGameState(STATE_COUNTDOWN);
	RevivePlayers();
	
	ChooseNewTrader();
	
	ACS_NamedExecute("WaveCountdown",0,0,0,0);
}

// ------------------------/
// Countdown script for the wave!
// ------------------------/

script "WaveCountdown" (void)
{
	SetGameTimer(COUNTDOWN_TIME);
	
	while (GameState.WaveTimer > 0)
	{
		// ZTAnnounce(StrParam(s:"WAVE ", d:GameState.Wave+1, s:" IS STARTING IN ", d:GameState.WaveTimer, s:" SECONDS..."));
		delay(35);
		SetGameTimer(GameState.WaveTimer-1);
	}
	
	BeginWaveProcess();
}

// ------------------------/
// Begin a wave's PROCESS, this is post-countdown!
// ------------------------/

void BeginWaveProcess()
{
	SetGameState(STATE_WAVE);
	GameState.ZombiesKilled = 0;
	
	bool isBoss = IsBossWave();
	
	// Normal wave, set appropriate goal and pool
	if (!isBoss)
	{
		// We want to KILL this many zombies
		GameState.ZombieGoal = ZOMBIEPOOL_BASE;
		// We can SPAWN this many zombies
		GameState.ZombiePool = ZOMBIEPOOL_BASE;
	}

	UpdateZombiesLeft();
	
	// Spawn director for normal zombies
	if (!isBoss)
	{
		ACS_NamedExecute("SpawnDirectorAI",0,0,0,0);
		PlayCombatSong();
	}
		
	// Spawn the actual boss!
	else
	{
		SpawnBossZombie();
		PlayBossSong();
	}
}

// ------------------------/
// Update the VISUAL number of zombies left for all players
// ------------------------/

void UpdateZombiesLeft()
{
	if (GameState.State == STATE_SANDBOX)
		return;
		
	Rep_UpdateCount();
	// All zombies have been killed!
	if (AllZombiesKilled())
		CompleteWave();
}

// ------------------------/
// Go to the next wave
// ------------------------/

void IncreaseWave()
{
	GameState.Wave ++;
}

// ------------------------/
// Complete a wave! All zombies were killed!
// ------------------------/

void CompleteWave()
{
	KillSpawnDirector();
	SetGameState(STATE_POSTWAVE);
	ACS_NamedExecuteAlways("PostWaveDelay",0,0,0,0);
}

// ------------------------/
// Do a slight delay before trader time
// ------------------------/

script "PostWaveDelay"
{
	Delay(35*2);
	BeginTraderTime();
}

// ------------------------/
// Actually begin trader time!
// ------------------------/

void BeginTraderTime()
{
	AmbientSound("music/trackchange", 127);
	PlayCalmSong();
	
	SetGameState(STATE_TRADER);
	PrintBold(s:"TRADER IS OPEN!!!!");
	
	RevivePlayers();
	
	OpenTraderDoor();
	
	SetGameTimer(TRADER_DEFAULTTIME);
	
	ACS_NamedExecute("TraderTimerScript",0,0,0,0);
}

// ------------------------/
// Responsible for controlling trader time
// ------------------------/

script "TraderTimerScript"
{
	while (GameState.WaveTimer > 0)
	{
		delay(35);
		SetGameTimer(GameState.WaveTimer-1);
	}
	
	CloseTraderDoor();
	TeleportPlayers();
	
	IncreaseWave();
	BeginWave();
}

// ------------------------/
// Open the trader doors
// ------------------------/

void OpenTraderDoor()
{
	int doorID = TraderArray[GameState.CurrentTrader].DoorID;
	Door_Open(doorID, 64);
}

// ------------------------/
// Close the trader doors
// ------------------------/

void CloseTraderDoor()
{
	int doorID = TraderArray[GameState.CurrentTrader].DoorID;
	Door_Close(doorID, 64);
}

// ------------------------/
// Enter / exit the trader
// ------------------------/

script SCRIPT_ENTERTRADER (int index) { EnterTrader(index); }
void EnterTrader(int tIndex)
{
	int PN = PlayerNumber();
	int LS = LineSide();
	
	// Enter from front
	if (LS == 0)
	{
		GiveInventory("InTrader", 1);
		PlayerState[PN].InTrader = true;
		PlayerState[PN].TraderIndex = tIndex;
	}
		
	// Exit from back
	else
	{
		TakeInventory("InTrader", 1);
		PlayerState[PN].InTrader = false;
		PlayerState[PN].TraderIndex = -1;
	}
}

// ------------------------/
// Teleport all players out of the trader!
// ------------------------/

using TElem: OpenTraderMenu;

void TeleportPlayers()
{
	int OID = ActivatorTID();
	int destSpot = TraderArray[GameState.CurrentTrader].SpotID;
	
	for (int i=0; i<MAXPLAYERS; i++)
	{
		// They're in the trader!
		if (PlayerState[i].InTrader && PlayerInGame(i))
		{
			SetActivator(i+BASE_PLAYER_TID);
			Teleport(destSpot, 0, 0);
			
			// Close the trader menu
			OpenTraderMenu(false);
		}
			
		TakeActorInventory( PlayerTID(i), "InTrader", 1);
		PlayerState[i].InTrader = false;
	}
	
	SetActivator(OID);
}

// ------------------------/
// Forcefully end the wave! 
// This is a debug command
// ------------------------/

script "ForceEndWave" (void) NET { ForceEndWave(); }
void ForceEndWave()
{
	SetActivator(-1);
	
	// Mid-wave? Kill all zombies
	if (GameState.State == STATE_WAVE)
	{
		PrintBold(s:"Forcefully ending the wave...");
		GameState.ZombiePool = 0;
		GameState.ZombieGoal = 0;
		GameState.ZombiesKilled = 100;
		KillZombies();
		
		UpdateZombiesLeft();
	}
	
	// Countdown, just increase
	else if (GameState.State == STATE_COUNTDOWN)
	{
		IncreaseWave();
		BeginWave();
	}
}

// -=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=